package com.arms.api.analysis.time.service;

import com.arms.api.analysis.time.dto.AlmIssueCountAndStatusByDateDTO;
import com.arms.api.analysis.time.dto.TimeAggrDTO;
import com.arms.api.analysis.time.dto.TimeDTO;
import com.arms.api.analysis.time.vo.AlmIssueStatByDateVO;
import com.arms.api.analysis.time.vo.RidgeLineVO;
import com.arms.api.issue.almapi.model.entity.AlmIssueEntity;
import com.arms.egovframework.javaservice.esframework.esquery.EsBoolQuery;
import com.arms.egovframework.javaservice.esframework.esquery.EsQuery;
import com.arms.egovframework.javaservice.esframework.esquery.EsQueryBuilder;
import com.arms.egovframework.javaservice.esframework.esquery.filter.RangeQueryFilter;
import com.arms.egovframework.javaservice.esframework.esquery.filter.TermQueryFilter;
import com.arms.egovframework.javaservice.esframework.esquery.filter.TermsQueryFilter;
import com.arms.egovframework.javaservice.esframework.esquery.must.TermQueryMust;
import com.arms.egovframework.javaservice.esframework.model.dto.집계_하위_필드;
import com.arms.egovframework.javaservice.esframework.model.vo.v2.DocumentAggregations;
import com.arms.egovframework.javaservice.esframework.model.vo.v2.DocumentBucket;
import com.arms.egovframework.javaservice.esframework.repository.EsCommonRepositoryWrapper;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Slf4j
@Service("analysisTime")
public class AnalysisTimeImpl extends EsCommonRepositoryWrapper<AlmIssueEntity> implements AnalysisTime {

    @Override
    public List<AlmIssueEntity> getUpdatedIssueForLastYear(TimeDTO timeDTO) {

        LocalDate now = LocalDate.now(ZoneId.of("UTC"));
        LocalDate startDate = now.minusYears(1L);

        EsQuery esQuery = new EsQueryBuilder()
                .bool(  new TermQueryMust("pdServiceId", timeDTO.getPdServiceLink()),
                        new TermsQueryFilter("pdServiceVersions", timeDTO.getPdServiceVersionLinks()),
                        RangeQueryFilter.of("updated").betweenDate(startDate.toString(),String.valueOf(now))
                );

        timeDTO.setEsQuery(esQuery);

        return findRecentHits(timeDTO).toDocs();
    }

    @Override
    public AlmIssueStatByDateVO getUpdatedIssueByDate(TimeAggrDTO timeAggrDTO) {

        timeAggrDTO.set메인_그룹_필드("updated");
        timeAggrDTO.set메인_그룹_필드명_별칭("updated");
        timeAggrDTO.set컨텐츠_보기_여부(true);

        timeAggrDTO.set집계_하위_필드들(
                List.of(
                        집계_하위_필드.builder()
                                .하위_필드명_별칭("isReq")
                                .하위_필드명("isReq")
                                .build()
                )
        );

        EsQuery esQuery = new EsQueryBuilder()
                .bool(
                        new TermQueryMust("pdServiceId", timeAggrDTO.getPdServiceLink()),
                        new TermsQueryFilter("pdServiceVersions", timeAggrDTO.getPdServiceVersionLinks()),
                        RangeQueryFilter.of("updated").betweenDate(timeAggrDTO.getStartDate(), timeAggrDTO.getEndDate())
                );

        timeAggrDTO.setEsQuery(esQuery);

        DocumentAggregations documentAggregations = aggregateDocsByDay(timeAggrDTO);
        List<DocumentBucket> documentBuckets = documentAggregations.deepestSubList();

        Map<String, AlmIssueCountAndStatusByDateDTO> issueStatByDate = documentBuckets.stream()
                .sorted(Comparator.comparing(bucket -> OffsetDateTime.parse(bucket.getDocFieldName("updated")).toLocalDate()))
                .collect(Collectors.toMap(
                        bucket -> transformDate(bucket.getDocFieldName("updated")), // 키: 업데이트 날짜
                        bucket -> {
                            // AlmIssueCountAndStatusByDateDTO 생성
                            AlmIssueCountAndStatusByDateDTO dto = new AlmIssueCountAndStatusByDateDTO();

                            // totalRequirements: isReq가 true인 경우
                            long totalRequirements = "true".equals(bucket.getDocFieldName("isReq"))
                                    ? bucket.getDocFieldNameCount("isReq")
                                    : 0L;
                            dto.setTotalRequirements(totalRequirements);

                            // totalRelationIssues: isReq가 false인 경우
                            long totalRelationIssues = "false".equals(bucket.getDocFieldName("isReq"))
                                    ? bucket.getDocFieldNameCount("isReq")
                                    : 0L;
                            dto.setTotalRelationIssues(totalRelationIssues);

                            return dto;
                        },
                        (existing, replacement) -> existing, // 중복 키 발생 시 기존 값 유지
                        LinkedHashMap::new
                ));

        return AlmIssueStatByDateVO.builder()
                .issueStatByDate(issueStatByDate)
                .build();
    }

    @Override
    public AlmIssueStatByDateVO getUpdatedIssueByDateAndStatus(TimeAggrDTO timeAggrDTO) {

        timeAggrDTO.set메인_그룹_필드("updated");
        timeAggrDTO.set메인_그룹_필드명_별칭("updated");
        timeAggrDTO.set컨텐츠_보기_여부(true);

        timeAggrDTO.set집계_하위_필드들(
                List.of(
                        집계_하위_필드.builder()
                                .하위_필드명_별칭("isReq")
                                .하위_필드명("isReq")
                                .build(),
                        집계_하위_필드.builder()
                                .하위_필드명_별칭("status")
                                .하위_필드명("status.status_name.keyword")
                                .build()
                )
        );

        EsQuery esQuery = new EsQueryBuilder()
                .bool(
                        new TermQueryMust("pdServiceId", timeAggrDTO.getPdServiceLink()),
                        new TermsQueryFilter("pdServiceVersions", timeAggrDTO.getPdServiceVersionLinks()),
                        RangeQueryFilter.of("updated").betweenDate(timeAggrDTO.getStartDate(), timeAggrDTO.getEndDate())
                );

        timeAggrDTO.setEsQuery(esQuery);

        DocumentAggregations documentAggregations = aggregateDocsByDay(timeAggrDTO);
        List<DocumentBucket> documentBuckets = documentAggregations.deepestSubList();

        Map<String, AlmIssueCountAndStatusByDateDTO> issueStatByDate = documentBuckets.stream()
                .sorted(Comparator.comparing(bucket -> OffsetDateTime.parse(bucket.getDocFieldName("updated")).toLocalDate()))
                .collect(Collectors.toMap(
                        bucket -> transformDate(bucket.getDocFieldName("updated")), // 키: 업데이트 날짜
                        bucket -> {
                            AlmIssueCountAndStatusByDateDTO dto = new AlmIssueCountAndStatusByDateDTO();

                            // 상태별 집계
                            Map<String, Long> requirementStatuses = new HashMap<>();
                            Map<String, Long> relationIssueStatuses = new HashMap<>();

                            String isReq = bucket.getDocFieldName("isReq");
                            String status = bucket.getDocFieldName("status");
                            long count = bucket.getDocFieldNameCount("status");

                            if ("true".equals(isReq)) {
                                requirementStatuses.put(status, count);
                            } else if ("false".equals(isReq)) {
                                relationIssueStatuses.put(status, count);
                            }

                            // 상태별 합계
                            dto.setRequirementStatuses(requirementStatuses);
                            dto.setRelationIssueStatuses(relationIssueStatuses);

                            // 총 요구사항과 연결 이슈 합계 계산
                            dto.setTotalRequirements(requirementStatuses.values().stream().mapToLong(Long::longValue).sum());
                            dto.setTotalRelationIssues(relationIssueStatuses.values().stream().mapToLong(Long::longValue).sum());

                            return dto;
                        },
                        (existing, replacement) -> {
                            // 중복 키 발생 시 상태별 데이터 병합
                            replacement.getRequirementStatuses().forEach(
                                    (key, value) -> existing.getRequirementStatuses().merge(key, value, Long::sum)
                            );
                            replacement.getRelationIssueStatuses().forEach(
                                    (key, value) -> existing.getRelationIssueStatuses().merge(key, value, Long::sum)
                            );

                            // 총 요구사항과 관계 이슈 값 업데이트
                            existing.setTotalRequirements(existing.getRequirementStatuses().values().stream().mapToLong(Long::longValue).sum());
                            existing.setTotalRelationIssues(existing.getRelationIssueStatuses().values().stream().mapToLong(Long::longValue).sum());

                            return existing;
                        },
                        LinkedHashMap::new
                ));

        return AlmIssueStatByDateVO.builder()
                .issueStatByDate(issueStatByDate)
                .build();

    }

    @Override
    public List<AlmIssueEntity> getUpdatedIssueByDateRange(TimeAggrDTO timeAggrDTO){

        return searchReqUpdatedDocument(timeAggrDTO);
    }

    @Override
    public List<RidgeLineVO> getUpdatedFrequency(TimeAggrDTO timeAggrDTO) {

        // 1. summary 조회를 위해 기간 내 요구사항 존재
        List<AlmIssueEntity> reqData = searchReqUpdatedDocument(timeAggrDTO);

        Map<String, String> keySummaryMap = reqData.stream()
                .filter(issue -> issue.getKey() != null && issue.getSummary() != null)
                .collect(Collectors.toMap(
                        AlmIssueEntity::getKey,
                        AlmIssueEntity::getSummary
                ));

        // 2. 요구사항의, 하위이슈 업데이트 횟수를 파악하기 위한 집계
        DocumentAggregations reqAggregateDocument= getAggregatedByUpdateField(timeAggrDTO, "key", true);
        DocumentAggregations subAggregateDocument = getAggregatedByUpdateField(timeAggrDTO, "parentReqKey", false);

        // 3. 집계한 데이터를 vo 형태로 반환
        List<RidgeLineVO> reqUpdateRidgeLineData = renderRidgeLineDate(reqAggregateDocument,true);
        List<RidgeLineVO> subUpdateRidgeLineData = renderRidgeLineDate(subAggregateDocument,false);

        // 4. 요규사항의 총 업데이트 수 = 하위이슈의 업데이트 수 + 요구사항의 업데이트 수, vo에 키에 맞는 서머리 set
        List<RidgeLineVO> ridgeLineData = calculateUpdatedDate(reqUpdateRidgeLineData, subUpdateRidgeLineData, keySummaryMap);

        // 5. 차트 데이터가 1개일 경우 차트에 보이는 형태가 이상해 앞 뒤 일에 값이 0인 padding data를 넣어줌
        return reRenderRidgeLineDate(ridgeLineData);
    }

    private List<AlmIssueEntity> searchReqUpdatedDocument(TimeAggrDTO timeAggrDTO){

        EsQuery esQuery = new EsQueryBuilder()
                .bool(
                        new TermQueryFilter("pdServiceId", timeAggrDTO.getPdServiceLink()),
                        new TermsQueryFilter("pdServiceVersions",timeAggrDTO.getPdServiceVersionLinks()),
                        new TermQueryFilter("isReq", true),
                        new RangeQueryFilter("updated", timeAggrDTO.getStartDate(), timeAggrDTO.getEndDate(), "fromto")
                );

        timeAggrDTO.setEsQuery(esQuery);

        return this.findRecentHits(timeAggrDTO).toDocs();
    }
    private DocumentAggregations getAggregatedByUpdateField(TimeAggrDTO timeAggrDTO , String subField, boolean isReq) {
        timeAggrDTO.set메인_그룹_필드("updated");
        timeAggrDTO.set메인_그룹_필드명_별칭("updated");

        timeAggrDTO.set집계_하위_필드들(
                List.of(
                        집계_하위_필드.builder()
                                .하위_필드명_별칭("pdServiceVersions")
                                .하위_필드명("pdServiceVersions")
                                .build(),
                        집계_하위_필드.builder()
                                .하위_필드명_별칭(subField)
                                .하위_필드명(subField)
                                .build()
                )
        );

        EsBoolQuery[] esBoolQueries = Stream.of(
                new TermQueryFilter("pdServiceId", timeAggrDTO.getPdServiceLink()),
                new TermsQueryFilter("pdServiceVersions",timeAggrDTO.getPdServiceVersionLinks()),
                new TermQueryFilter("isReq", isReq),
                new RangeQueryFilter("updated", timeAggrDTO.getStartDate(), timeAggrDTO.getEndDate(), "fromto")
        ).toArray(EsBoolQuery[]::new);

        EsQuery esQuery = new EsQueryBuilder().bool(esBoolQueries);
        timeAggrDTO.setEsQuery(esQuery);

        return this.aggregateDocsByDay(timeAggrDTO);
    }

    private List<RidgeLineVO> renderRidgeLineDate(DocumentAggregations aggregatedData,boolean isReq){

        List<DocumentBucket> documentBuckets = aggregatedData.deepestSubList();

        List<RidgeLineVO> collect = documentBuckets.stream().map(
                        a -> {
                            return RidgeLineVO.builder()
                                    .updateDate(transformDate(a.getDocFieldName("updated")))
                                    .version(a.getDocFieldName("pdServiceVersions"))
                                    .key(isReq ? a.getDocFieldName("key") : a.getDocFieldName("parentReqKey"))
                                    .build();
                        }
                )
                .peek(a-> System.out.println(a.getKey()))
                .collect(Collectors.toList());

        return collect;
    }
    private String transformDate(String date) {
        String subDate = date.substring(0,10);
        LocalDate localDate = LocalDate.parse(subDate);
        return DateTimeFormatter.ofPattern("yyyy-MM-dd").format(localDate);
    }

    private List<RidgeLineVO> calculateUpdatedDate(List<RidgeLineVO> renderedReqDate, List<RidgeLineVO> renderedSubDate ,Map<String, String> keySummaryMap ){

        List<RidgeLineVO> mergedList = new ArrayList<>(renderedReqDate);

        mergedList.addAll(renderedSubDate);

        // 버전, 업데이트일, 키로 그룹화
        Map<RidgeLineVO, Long> groupedMap = mergedList.stream()
                .filter(vo -> vo.getKey() != null && !vo.getKey().isEmpty())
                .collect(Collectors.groupingBy(
                        vo -> RidgeLineVO.builder()
                                .version(vo.getVersion())
                                .updateDate(vo.getUpdateDate())
                                .key(vo.getKey())
                                .build(),
                        Collectors.counting()
                ));

        return groupedMap.entrySet().stream()
                .map(entry -> RidgeLineVO.builder()
                        .version(entry.getKey().getVersion())
                        .updateDate(entry.getKey().getUpdateDate())
                        .key(entry.getKey().getKey())
                        .updatedCount(entry.getValue())
                        .summary(keySummaryMap.get(entry.getKey().getKey()))
                        .build()
                )
                .collect(Collectors.toList());
    }

    private List<RidgeLineVO> reRenderRidgeLineDate(List<RidgeLineVO> ridgeLineData) {

        if (ridgeLineData.size() == 1) {
            RidgeLineVO singleData = ridgeLineData.get(0);

            LocalDate currentDate = LocalDate.parse(singleData.getUpdateDate());

            String previousDate = currentDate.minusDays(1).toString();
            String nextDate = currentDate.plusDays(1).toString();

            RidgeLineVO previousDayData = RidgeLineVO.builder()
                    .version(singleData.getVersion())
                    .updateDate(previousDate)
                    .key(singleData.getKey())
                    .updatedCount(0)
                    .summary(singleData.getSummary())
                    .build();

            RidgeLineVO nextDayData = RidgeLineVO.builder()
                    .version(singleData.getVersion())
                    .updateDate(nextDate)
                    .key(singleData.getKey())
                    .updatedCount(0)
                    .summary(singleData.getSummary())
                    .build();

            return List.of(previousDayData, singleData, nextDayData);
        }
        return ridgeLineData;
    }
}

